#!/usr/bin/env python
"""
This script visualizes results with external graphical applications.
or convert data from Abinit files (usually netcdf) to other formats.
"""
import sys
import os
import argparse

from pprint import pformat
from monty.functools import prof_main
from monty.termcolor import cprint
from abipy import abilab
from abipy.iotools.visualizer import Visualizer
from abipy.tools.plotting import MplExpose, PanelExpose, GenericDataFilePlotter, plotlyfigs_to_browser, push_to_chart_studio


def handle_overwrite(path, options):
    """Exit 1 if file ``path`` exists and not options.force else return path."""
    name_parts = os.path.splitext(path)
    print("Writing %s file:" % name_parts[-1].replace(".", "").upper())
    if os.path.exists(path) and not options.force:
        cprint("Cannot overwrite pre-existent file. Use `-f` options.", "red")
        sys.exit(1)
    return path


def df_to_clipboard(options, df):
    """Copy dataframe to clipboard if options.clipboard."""
    if getattr(options, "clipboard", False):
        cprint("Copying dataframe to the system clipboard.", "green")
        cprint("This can be pasted into Excel, for example", "green")
        df.to_clipboard()


class NegateAction(argparse.Action):
    def __call__(self, parser, ns, values, option):
        setattr(ns, self.dest, option[2:4] != 'no')


#def abiview_fields(options):
#    """Animate fields with Mayavi. Accept any file with density or potential ..."""
#    from abipy.display.mvtk import MayaviFieldAnimator
#    a = MayaviFieldAnimator(options.filepath)
#    a.volume_animate()
#    return 0


def abiview_structure(options):
    """
    Visualize the structure with the specified visualizer. Requires external app
    or optional python modules (mayavi, vtk)
    """
    structure = abilab.Structure.from_file(options.filepath)
    print(structure.to_string(verbose=options.verbose))
    print("Visualizing structure with:", options.appname)
    structure.visualize(appname=options.appname)
    return 0


def abiview_hist(options):
    """
    Visualize structural relaxation/molecular-dynamics run
    from data stored in the HIST.nc file. Requires mayavi.
    """
    with abilab.abiopen(options.filepath) as hist:
        print(hist.to_string(verbose=options.verbose))
        if options.appname is not None:
            hist.visualize(appname=options.appname, to_unit_cell=options.to_unit_cell)
        elif options.xdatcar:
            xpath = options.filepath + ".XDATCAR"
            handle_overwrite(xpath, options)
            hist.write_xdatcar(filepath=xpath, groupby_type=True, overwrite=True,
                               to_unit_cell=options.to_unit_cell)
        else:
            hist.plot()

        return 0


def abiview_data(options):
    """
    Extract data from a generic text file with results
    in tabular format and plot data with matplotlib.
    Multiple datasets are supported.
    No attempt is made to handle metadata (e.g. column name)
    """
    plotter = GenericDataFilePlotter(options.filepath)
    print(plotter.to_string(verbose=options.verbose))
    plotter.plot(use_index=options.use_index)

    return 0


#def abiview_abo(options):
#    """
#    Plot self-consistent iterations extracted from Abinit output file
#    as well as timer data (if present)
#    """
#    with abilab.abiopen(options.filepath) as abo:
#        print(abo.to_string(verbose=options.verbose))
#        abo.plot()
#        #timer = abo.get_timer()
#    return 0


def abiview_dirviz(options):
    """
    Visualize directory tree with graphviz.
    """
    from pymatgen.io.abinit.utils import Dirviz
    import tempfile
    graph = Dirviz(options.filepath).get_cluster_graph(engine=options.engine)
    directory = tempfile.mkdtemp()
    print("Producing source files in:", directory)
    graph.view(directory=directory)

    return 0


def abiview_ebands(options):
    """
    Plot electronic bands if file contains high-symmetry k-path or DOS if k-sampling.
    Accept any file with ElectronBands e.g. GSR.nc, WFK.nc, ...
    """
    with abilab.abiopen(options.filepath) as abifile:
        if options.xmgrace:
            outpath = options.filepath + ".agr"
            abifile.ebands.to_xmgrace(handle_overwrite(outpath, options))
        elif options.bxsf:
            outpath = options.filepath + ".bxsf"
            abifile.ebands.to_bxsf(handle_overwrite(outpath, options))
        #elif options.plotly:
        #    print(abifile.to_string(verbose=options.verbose))
        else:
            print(abifile.to_string(verbose=options.verbose))
            abifile.expose_ebands(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                                  expose_web=options.expose_web, verbose=options.verbose)

        return 0


def abiview_skw(options):
    """
    Interpolate energies in k-space along a k-path with star-function methods
    Note that the interpolation will likely fail if there are symmetrical k-points in the input set of k-points
    so it's recommended to call this method with energies obtained in the IBZ.
    Accept any file with ElectronBands e.g. GSR.nc, WFK.nc, ...
    """
    ebands = abilab.ElectronBands.as_ebands(options.filepath)
    if not ebands.kpoints.is_ibz:
        cprint("SKW interpolator should be called with energies in the IBZ", "yellow")
    r = ebands.interpolate(lpratio=options.lpratio, line_density=options.line_density, verbose=options.verbose)
    r.ebands_kpath.plot()
    return 0


def abiview_fs(options):
    """
    Extract eigenvalues in the IBZ from file and visualize Fermi surface with
    the external application specified via --appname
    """
    with abilab.abiopen(options.filepath) as abifile:
        eb3d = abifile.ebands.get_ebands3d()

        if options.appname == "mpl":
            eb3d.plot_isosurfaces()
        elif options.appname == "xsf":
            eb3d.xcrysden_view()
        elif options.appname == "mayavi":
            eb3d.mvplot_isosurfaces()
        else:
            raise ValueError("Unsupported value for --appname: %s" % str(options.appname))

        return 0


def abiview_ifermi_fs(options):
    """
    Use ifermi package to visualize the Fermi surface. Requires netcdf file with energies in the IBZ.
    See <https://fermisurfaces.github.io/IFermi/>
    """
    with abilab.abiopen(options.filepath) as abifile:

        kwargs = dict(
            interpolation_factor=options.interpolation_factor,
            mu=options.mu,
            eref=options.eref,
            wigner_seitz=options.wigner,
            with_velocities=options.with_velocities,
        )
        print(abifile.ebands, "\n")
        print("Building Fermi surface with options:\n\n", pformat(kwargs, indent=4), end=2*"\n")

        r = abifile.ebands.get_ifermi_fs(**kwargs)
        r.fs_plotter.get_plot(plot_type=options.plot_type).show()

        #abifile.ebands.get_ifermi_slices(**kwargs)

        return 0


def abiview_ddb(options):
    """
    Invoke Anaddb to compute phonon bands and DOS from the DDB, plot the results.
    """
    with abilab.abiopen(options.filepath) as ddb:
        print(ddb.to_string(verbose=options.verbose))

        # Don't need PHDOS if phononwebsite
        nqsmall = 0 if options.phononwebsite else 10
        ndivsm = 20; asr = 2; chneut = 1; dipdip = 1; dos_method = "tetra"; lo_to_splitting = "automatic"
        print("""
Computing phonon bands and DOS from DDB file with:
nqsmall: {nqsmall}, ndivsm: {ndivsm},
asr: {asr}, chneut: {chneut}, dipdip: {dipdip}, lo_to_splitting: {lo_to_splitting}, dos_method: {dos_method}
""".format(**locals()))

        print("Invoking anaddb ...  ", end="")
        phbst_file, phdos_file = ddb.anaget_phbst_and_phdos_files(
            nqsmall=nqsmall, ndivsm=ndivsm, asr=asr, chneut=chneut, dipdip=dipdip, dos_method=dos_method,
            lo_to_splitting=lo_to_splitting, verbose=options.verbose, mpi_procs=1)
        print("Calculation completed.\nResults available in:", os.path.dirname(phbst_file.filepath))

        phbands = phbst_file.phbands
        units = "mev"

        if options.xmgrace:
            outpath = options.filepath + ".agr"
            phbands.to_xmgrace(handle_overwrite(outpath, options))
            return 0

        #elif options.bxsf:
        #    outpath = options.filepath + ".bxsf"
        #    phbands.to_bxsf(handle_overwrite(outpath, options))
        #    return 0

        elif options.phononwebsite:
            return phbands.view_phononwebsite(browser=options.browser, verbose=options.verbose)

        elif options.plotly:
            # Plotly + Panel version.
            phdos = phdos_file.phdos
            with PanelExpose(title=f"Vibrational properties of {phdos_file.structure.formula}") as e:
                e(phbands.qpoints.plotly(show=False))
                e(phbands.plotly_with_phdos(phdos, units=units, show=False))
                e(phdos_file.plotly_pjdos_type(units=units, show=False))
                e(phdos.plotly_harmonic_thermo(tstart=5, tstop=300, num=50, units="eV", formula_units=1,
                                               quantities="all", show=False))
                e(phdos.plotly(units=units, show=False))
                e(phbands.plot_colored_matched(units=units, show=False))
                e(phbands.plot_fatbands(units=units, phdos_file=phdos_file, show=False))
                #push_to_chart_studio(figs) if options.chart_studio else plotlyfigs_to_browser(figs)

        else:
            phdos = phdos_file.phdos

            if not options.expose_web:
                # matplotlib figure and X-server.
                e = MplExpose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout)
            else:
                # panel version with matplotlib.
                e = PanelExpose(title=f"Vibrational properties of {phdos_file.structure.formula}")

            with e:
                e(phbands.qpoints.plot(show=False))
                e(phbands.plot_with_phdos(phdos, units=units, show=False))
                e(phdos_file.plot_pjdos_type(units=units, show=False))
                e(phbands.plot_colored_matched(units=units, show=False))
                e(phdos.plot_harmonic_thermo(tstart=5, tstop=300, num=50, units="eV", formula_units=1,
                             quantities="all", show=False))
                e(phdos.plot(units=units, show=False))
                e(phbands.plot_fatbands(units=units, phdos_file=phdos_file, show=False))

        phbst_file.close()
        phdos_file.close()

    return 0


def abiview_ddb_vs(options):
    """
    Compute speed of sound by fitting phonon frequencies along selected directions.
    """
    num_points = 20; asr = 2; chneut = 1; dipdip = 1
    print("""
Computing phonon frequencies for linear least-squares with:
num_points: {num_points}, asr: {asr}, chneut: {chneut}, dipdip: {dipdip}
""".format(**locals()))

    print("Invoking anaddb ...  ")
    sv = abilab.SoundVelocity.from_ddb(options.filepath, num_points=num_points,
                                       asr=asr, chneut=chneut, dipdip=dipdip, verbose=options.verbose)
    #print("Calculation completed.\nResults available in:", os.path.dirname(phbst_file.filepath))

    df = sv.get_dataframe()
    abilab.print_dataframe(df, title="Speed of sound for different directions:")
    df_to_clipboard(options, df)
    sv.plotly if options.plotly else sv.plot()

    return 0


def abiview_ddb_ir(options):
    """
    Compute infra-red spectrum from DDB. Plot results.
    """
    asr = 2; chneut = 1; dipdip = 1
    print("""
Computing phonon frequencies for infra-red spectrum with:
asr: {asr}, chneut: {chneut}, dipdip: {dipdip}
""".format(**locals()))

    with abilab.abiopen(options.filepath) as ddb:
        tgen = ddb.anaget_dielectric_tensor_generator(asr=asr, chneut=chneut, dipdip=dipdip, verbose=options.verbose)
        print(tgen.to_string(verbose=options.verbose))

        gamma_ev = 1e-4
        print("Plotting dielectric tensor with constant phonon damping: %s (eV)" % gamma_ev)
        if options.plotly:
            tgen.plotly_all(gamma_ev=gamma_ev)
        else:
            tgen.plot_all(gamma_ev=gamma_ev)

    return 0


def abiview_ddb_asr(options):
    """
    Compute phonon band structure from DDB with/without acoustic sum rule. Plot results.
    """
    print("Computing phonon frequencies with/without ASR")

    with abilab.abiopen(options.filepath) as ddb:
        plotter = ddb.anacompare_asr(asr_list=(0, 2), chneut_list=(1,), dipdip=1, lo_to_splitting="automatic",
                                     nqsmall=10, ndivsm=20, dos_method="tetra", ngqpt=None,
                                     verbose=0, mpi_procs=1)

        title = ddb.structure.formula
        renderer = "browser" if not options.chart_studio else "chart_studio"
        plotter.combiplotly(renderer=renderer, title=title) if options.plotly else plotter.plot(title=title)

    return 0


def abiview_ddb_dipdip(options):
    """
    Compute phonon band structure from DDB with/without dipole-dipole interaction. Plot results.
    """
    print("Computing phonon frequencies with/without dipdip")

    with abilab.abiopen(options.filepath) as ddb:
        plotter = ddb.anacompare_dipdip(chneut_list=(1,), asr=2, lo_to_splitting="automatic",
                                        nqsmall=10, ndivsm=20, dos_method="tetra", ngqpt=None,
                                        verbose=0, mpi_procs=1)

        title = ddb.structure.formula
        renderer = "browser" if not options.chart_studio else "chart_studio"
        plotter.combiplotly(renderer=renderer, title=title) if options.plotly else plotter.plot(title=title)

    return 0


def abiview_ddb_quad(options):
    """
    Compute phonon band structure from DDB with/without quadrupole terms. Plot results.
    """
    print("Computing phonon frequencies with/without dip-quad and quad-quad terms.")

    with abilab.abiopen(options.filepath) as ddb:
        plotter = ddb.anacompare_quad(asr=2, chneut=1, dipdip=1, lo_to_splitting="automatic",
                                      nqsmall=0, ndivsm=20, dos_method="tetra", ngqpt=None,
                                      verbose=0, mpi_procs=1)

        title = ddb.structure.formula
        renderer = "browser" if not options.chart_studio else "chart_studio"
        plotter.combiplotly(renderer=renderer, title=title) if options.plotly else plotter.plot(title=title)

    return 0


def abiview_ddb_isodistort_ph(options):
    """
    Compute ph-freqs for given q-point (default: Gamma), produce CIF files for unperturbed and distorded structure
    that can be used with ISODISTORT (https://stokes.byu.edu/iso/isodistort.php) to analyze the symmetry of phonon modes.
    See README.me file produced in output directory.
    """
    with abilab.abiopen(options.filepath) as ddb:
        print(ddb.to_string(verbose=options.verbose))
        qpoint = options.qpoint
        asr = 2; chneut = 1; dipdip = 1; lo_to_splitting = False
        print("""
Computing phonon frequencies and eigenvectors with:
asr: {asr}, chneut: {chneut}, dipdip: {dipdip}, lo_to_splitting: {lo_to_splitting}
qpoint = {qpoint}
""".format(**locals()))

        print("Invoking anaddb ...  ", end="")
        phbands = ddb.anaget_phmodes_at_qpoint(qpoint=qpoint, asr=asr, chneut=chneut, dipdip=dipdip,
                                               verbose=options.verbose, lo_to_splitting=False)

        phbands.make_isodistort_ph_dir(qpoint, select_modes=None, eta=1, workdir=None)

    return 0


def abiview_ddb_qpt(options):
    """
    Compute ph-frequencies at the selected q-point without passing through the
    Fourier interpolation of the interatomic force constants.
    """
    #    asr = 2; chneut = 1; dipdip = 1
    #    print("""
    #Computing interatomic force constants with
    #asr: {asr}, chneut: {chneut}, dipdip: {dipdip}
    #""".format(**locals()))

    with abilab.abiopen(options.filepath) as ddb:
        # Execute anaddb to compute phbands without Fourier interpolation.
        phbands, inp = ddb.anaget_phmodes_at_qpoint(qpoint=options.qpoint,
                            asr=2, chneut=1, dipdip=1, ifcflag=0, return_input=True)
        print(inp)
        df = phbands.get_dataframe()
        abilab.print_dataframe(df, title="Phonon frequencies:")
        df_to_clipboard(options, df)

    return 0


def abiview_ddb_ifc(options):
    """
    Visualize interatomic force constants in real space.
    """
    asr = 2; chneut = 1; dipdip = 1
    print("""
Computing interatomic force constants with
asr: {asr}, chneut: {chneut}, dipdip: {dipdip}
""".format(**locals()))

    with abilab.abiopen(options.filepath) as ddb:
        # Execute anaddb to compute the interatomic force constants.
        ifc = ddb.anaget_ifc(asr=asr, chneut=chneut, dipdip=dipdip)
        #print(ifc)

        if not options.expose_web:
            # matplotlib figure and X-server.
            e = MplExpose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout)
        else:
            # panel version
            e = PanelExpose(title=f"Interatomic force constants of {phdos_file.structure.formula}")

        with e:
            e(ifc.plot_longitudinal_ifc(title="Longitudinal IFCs", show=False))
            e(ifc.plot_longitudinal_ifc_short_range(title="Longitudinal IFCs short range", show=False))
            e(ifc.plot_longitudinal_ifc_ewald(title="Longitudinal IFCs Ewald", show=False))

    return 0


def abiview_phbands(options):
    """
    Plot phonon bands. Accept any file with PhononBands e.g. PHBST.nc, ...
    """
    with abilab.abiopen(options.filepath) as abifile:
        if options.xmgrace:
            outpath = options.filepath + ".agr"
            abifile.phbands.to_xmgrace(handle_overwrite(outpath, options))
        #elif options.bxsf:
        #    outpath = options.filepath + ".bxsf"
        #    abifile.phbands.to_bxsf(handle_overwrite(outpath, options))
        #    return 0
        elif options.phononwebsite:
            return abifile.phbands.view_phononwebsite(browser=options.browser)
        else:
            print(abifile.to_string(verbose=options.verbose))
            abifile.expose_phbands(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                           verbose=options.verbose, units="mev")

        return 0


def abiview_denpot(options):
    """
    Plot DEN/POT files
    """
    with abilab.abiopen(options.filepath) as abifile:
        print(abifile.to_string(verbose=options.verbose))
        field = abifile.field
        if options.chgcar:
            if not field.is_density_like:
                cprint("Warning: expecting DENSITY file but received %s" % field.__class__.__name__, "yellow")
            chg_path = options.filepath + ".CHGCAR"
            field.to_chgcar(filename=handle_overwrite(chg_path, options))
        #elif options.cube:
        #    outpath = options.filepath + ".cube"
        #    field.export_to_cube(filename=handle_overwrite(outpath, options), spin="total")
        else:
            abifile.field.visualize(appname=options.appname)
    return 0


def abiview_lobster(options):
    """Analyze lobster output files in directory."""
    from abipy.electrons.lobster import LobsterAnalyzer
    lobana = LobsterAnalyzer.from_dir(os.path.dirname(options.filepath), prefix=options.prefix)
    print(lobana.to_string(verbose=options.verbose))

    if options.ipython:
        # Start ipython shell with namespace
        # Use embed because I don't know how to show a header with start_ipython.
        import IPython
        IPython.embed(header="The LobsterAnalyzer is bound to the `lobana` variable.\nTry `print(lobana)`")

    elif options.notebook:
        return lobana.make_and_open_notebook(foreground=options.foreground)

    else:
        lobana.plot()
        #lobana.plot_coxp_with_dos(from_site_index=[0, 1])

    return 0


def get_epilog():
    return """\
Usage example:

###########
# Structure
###########

    abiview.py structure FILE                ==> Visualize structure with Vesta (default)
    abiview.py structure FILE -a xcrysden    ==> Visualize structure with Xcrysden (default)
    abiview.py hist out_HIST.nc              ==> Plot relaxation/molecular dynamics results with matplotlib.
    abiview.py hist out_HIST.nc -a ovito     ==> Visualize relaxation/molecular dynamics results with ovito.
    abiview.py hist out_HIST.nc --xdatcar    ==> Convert HIST.nc into XDATCAR format (caveat: assume fixed unit cell!)

############
# Text files
############

    abiview.py data FILE     ==> Parse text FILE with data in tabular format and plot arrays.

###########
# Electrons
###########

    abiview.py ebands out_WFK.nc              ==>  Plot electrons bands (or DOS) with matplotlib.
    abiview.py ebands out_GSR.nc --xmgrace    ==>  Generate xmgrace file with electron bands.
    abiview.py fs FILE_WITH_KMESH.nc          ==>  Visualize Fermi surface from netcdf file with electron energies
                                                   on a k-mesh. Use -a xsf to change application e.g. Xcrysden.
    abiview.py skw out_GSR.nc                 ==> Interpolate IBZ energies with star-functions and plot
                                                  interpolated bands.

#########
# Phonons
#########

    abiview.py ddb in_DDB                 ==>  Compute ph-bands and DOS from DDB, plot results.
    abiview.py ddb_vs                     ==>  Compute speed of sound from DDB by fitting phonon frequencies.
    abiview.py ddb_ir                     ==>  Compute infra-red spectrum from DDB. Plot results.
    abiview.py ddb_asr                    ==>  Compute ph-bands from DDB with/wo acoustic rule.
                                               Plot results with matplotlib (default) or plotly (--plotly)
    abiview.py ddb_asr --plotly -cs       ==>  Compute ph-bands from DDB with/wo acoustic rule.
                                               Plot results with plotly and push figure
                                               to plotly chart studio cloud. See: https://chart-studio.plotly.com
    abiview.py ddb_dipdip                 ==>  Compute ph-bands from DDB with/wo dipole-dipole treatment.
                                               Plot results with matplotlib (default) or plotly (--plotly)
    abiview.py ddb_quad                   ==>  Compute ph-bands from DDB with/wo dipole-quadrupole terms.
                                               Plot results with matplotlib (default) or plotly (--plotly)
    abiview.py ddb_qpt -q 0 0.5 0         ==>  Compute ph-frequencies at the selected q-point without passing
                                               through the Fourier interpolation of the interatomic force constants.
    abiview.py ddb_ifc                    ==>  Visualize interatomic force constants in real space.
    abiview.py phbands out_PHBST.nc -web  ==>  Visualize ph-bands and displacements with phononwebsite.

###############
# Miscelleanous
###############

  abiview.py dirviz DIRECTORY            ==> Visualize directory tree with graphviz.
  abiview.py lobster DIRECTORY           ==> Visualize Lobster results.

Use `abiview.py --help` for help and `abiview.py COMMAND --help` to get the documentation for `COMMAND`.
Use `-v` to increase verbosity level (can be supplied multiple times e.g -vv).
"""

# TODO
#abiview.py denpot out_DEN.nc              ==>  Visualize density with Vesta.
#abiview.py denpot out_DEN.nc --chgcar     ==>  Convert DEN file into CHGCAR fileformat.


def get_parser(with_epilog=False):

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('filepath', type=str, help="File to visualize.")

    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity.')
    copts_parser.add_argument('-sns', "--seaborn", const="paper", default=None, action='store', nargs='?', type=str,
        help='Use seaborn settings. Accept value defining context in ("paper", "notebook", "talk", "poster"). Default: paper')
    copts_parser.add_argument('-mpl', "--mpl-backend", default=None,
        help=("Set matplotlib interactive backend. "
              "Possible values: GTKAgg, GTK3Agg, GTK, GTKCairo, GTK3Cairo, WXAgg, WX, TkAgg, Qt4Agg, Qt5Agg, macosx."
              "See also: https://matplotlib.org/faq/usage_faq.html#what-is-a-backend."))

    # Parent parser for commands supporting MplExpose.
    slide_parser = argparse.ArgumentParser(add_help=False)
    slide_parser.add_argument("-s", "--slide-mode", default=False, action="store_true",
            help="Iterate over figures. Expose all figures at once if not given on the CLI.")
    slide_parser.add_argument("-t", "--slide-timeout", type=int, default=None,
            help="Close figure after slide-timeout seconds (only if slide-mode). Block if not specified.")
    slide_parser.add_argument("-ew", "--expose-web", default=False, action="store_true",
            help='Generate matplotlib plots in $BROWSER instead of X-server. WARNING: Not all the features are supported.')

    # Parent parser for commands that operating on pandas dataframes
    pandas_parser = argparse.ArgumentParser(add_help=False)
    pandas_parser.add_argument("-c", '--clipboard', default=False, action="store_true",
            help="Copy dataframe to the system clipboard. This can be pasted into Excel, for example")

    # Parent parser for commands supporting ipython
    ipy_parser = argparse.ArgumentParser(add_help=False)
    ipy_parser.add_argument('-ipy', '--ipython', default=False, action="store_true", help='Invoke ipython terminal.')

    # Parent parser for commands supporting (jupyter notebooks)
    nb_parser = argparse.ArgumentParser(add_help=False)
    nb_parser.add_argument('-nb', '--notebook', default=False, action="store_true", help='Generate jupyter notebook.')
    nb_parser.add_argument('--foreground', action='store_true', default=False,
        help="Run jupyter notebook in the foreground.")

    # Parent parser for commands supporting plotly plots
    plotly_parser = argparse.ArgumentParser(add_help=False)
    plotly_parser.add_argument("-ply", '--plotly', default=False, action="store_true",
        help='Generate plotly plots in browser instead of matplotlib.')
    plotly_parser.add_argument("-cs", "--chart-studio", default=False, action="store_true",
        help="Push figure to plotly chart studio. " +
             "Requires --plotly and user account at https://chart-studio.plotly.com.")

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)

    # Create the parsers for the sub-command
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    def add_args(p, *args):
        """Add arguments to subparser `p`."""
        for arg in args:
            if arg == "xmgrace":
                p.add_argument('--xmgrace', default=False, action="store_true",
                    help="Print bands in xmgrace format to stdout and exit.")
            elif arg == "bxsf":
                p.add_argument('--bxsf', default=False, action="store_true",
                    help=("Generate BXSF file suitable for the visualization of isosurfaces with Xcrysden"
                          "(xcrysden --bxsf FILE).\n Requires k-points in IBZ. Print to stdout and exit."))
            elif arg == "phononweb":
                p.add_argument("-web", "--phononwebsite", default=False, action="store_true",
                    help=("Visualize phonon band structure on the phononwebsite. "
                          "http://henriquemiranda.github.io/phononwebsite/"))
            elif arg == "browser":
                p.add_argument("-b", "--browser", default=None,
                    help="Define browser used by python webbrowser. "
                         "See https://docs.python.org/2/library/webbrowser.html#webbrowser.register")
            elif arg == "force":
                p.add_argument("-f", '--force', default=False, action="store_true",
                    help="Overwrite pre-existent files without prompting for confirmation.")
            else:
                raise ValueError("Invalid arg: %s" % arg)

    # Subparser for structure command.
    p_structure = subparsers.add_parser('structure', parents=[copts_parser], help=abiview_structure.__doc__)
    p_structure.add_argument("-a", "--appname", nargs="?", type=str, default="vesta",
        help=("Application name. Default: vesta. "
              "Possible options: %s, mayavi, vtk" % ", ".join(Visualizer.all_visunames())))

    # Subparser for hist command.
    p_hist = subparsers.add_parser('hist', parents=[copts_parser], help=abiview_hist.__doc__)
    p_hist.add_argument("-a", "--appname", nargs="?", default=None, const="ovito",
        help=("Application name. Default: ovito. "
              "Possible options: `%s`, `mpl` (matplotlib) `mayavi`, `vtk`" % ", ".join(Visualizer.all_visunames())))
    p_hist.add_argument("--xdatcar", default=False, action="store_true", help="Convert HIST file into XDATCAR format.")
    p_hist.add_argument("--to-unit-cell", default=False, action="store_true",
            help="Whether to translate sites into the unit cell.")
    add_args(p_hist, "force")

    # Subparser for data command.
    p_data = subparsers.add_parser('data', parents=[copts_parser], help=abiview_data.__doc__)
    p_data.add_argument("-i", "--use-index", default=False, action="store_true",
        help="Use the row index as x-value in the plot. By default the plotter uses the first column as x-values")

    # Subparser for abo command.
    #p_abo = subparsers.add_parser('abo', parents=[copts_parser], help=abiview_abo.__doc__)

    # Subparser for dirviz command.
    p_dirviz = subparsers.add_parser('dirviz', parents=[copts_parser], help=abiview_dirviz.__doc__)
    p_dirviz.add_argument("-e", "--engine", type=str, default="fdp",
        help=("graphviz engine: ['dot', 'neato', 'twopi', 'circo', 'fdp', 'sfdp', 'patchwork', 'osage']. "
            "See http://www.graphviz.org/pdf/dot.1.pdf "
            "Use `conda install python-graphviz` or `pip install graphviz` to install the python package."))

    # Subparser for ebands command.
    p_ebands = subparsers.add_parser('ebands', parents=[copts_parser, slide_parser], help=abiview_ebands.__doc__)
    add_args(p_ebands, "xmgrace", "bxsf", "force")

    # Subparser for ebands command.
    p_skw = subparsers.add_parser('skw', parents=[copts_parser], help=abiview_skw.__doc__)
    p_skw.add_argument("-lp", "--lpratio", type=int, default=5,
        help=("Ratio between the number of star functions and the number of ab-initio k-points. "
              "The default should be OK in many systems, larger values may be required for accurate derivatives."))
    p_skw.add_argument("-ld", "--line-density", type=int, default=20,
                      help="Number of points in the smallest segment of the k-path.")

    # Subparser for fs command.
    p_fs = subparsers.add_parser('fs', parents=[copts_parser], help=abiview_fs.__doc__)
    p_fs.add_argument("-a", "--appname", type=str, default="mpl",
        help="Application name. Possible options: mpl (matplotlib, default), xsf (xcrysden), mayavi.")

    # Parent parser for ifermi commands
    ifermi_parser = argparse.ArgumentParser(add_help=False)
    ifermi_parser.add_argument("-i", '--interpolation-factor', default=8, type=float,
        help="interpolation factor for band structure  [default: 8.0]")
    ifermi_parser.add_argument("--eref", default="fermie", type=str, choices=['fermie', 'cbm', 'vbm'],
        help="Energy reference for isosurface. `eref` and `mu` define the energy level: isoe = eref + mu"  +
              "Use `fermie` for metals, `cbm` for the conduction band minimum, and `vbm` for the valence band maximum, " +
             "[default: `fermie`]")
    ifermi_parser.add_argument("-t", '--plot-type', default="plotly", choices=["plotly", "matplotlib", "mayavi"],
        help="Plot type. Possible options: plotly (default), matplotlib, mayavi.")
    ifermi_parser.add_argument('--wigner', '--no-wigner', dest='wigner', default=True, action=NegateAction, nargs=0,
            help="Use the Wigner-Seitz cell or the reciprocal lattice parallelepiped. Default is: wigner")
    ifermi_parser.add_argument("-vel", '--with-velocities', default=False, action="store_true",
            help="Show color map with interpolated group velocities. Default is False.")

    # Subparser for ifermi_fs command.
    p_ifermi_fs = subparsers.add_parser('ifermi_fs', parents=[ifermi_parser, copts_parser], help=abiview_ifermi_fs.__doc__)
    p_ifermi_fs.add_argument("-m", '--mu', default=0.0, type=float,
        help="Offset in eV from the energy reference, eref, at which to calculate the isosurface. Default 0 i.e. use `eref`")

    # Subparser for ddb command.
    p_ddb = subparsers.add_parser('ddb', parents=[copts_parser, slide_parser, plotly_parser], help=abiview_ddb.__doc__)
    add_args(p_ddb, "xmgrace", "phononweb", "browser", "force")

    # Subparser for ddb_vs command.
    p_ddb_vs = subparsers.add_parser('ddb_vs', parents=[copts_parser, pandas_parser, slide_parser],
                                     help=abiview_ddb_vs.__doc__)

    # Subparser for ddb_ir command.
    p_ddb_ir = subparsers.add_parser('ddb_ir', parents=[copts_parser, pandas_parser, slide_parser],
                                     help=abiview_ddb_ir.__doc__)

    # Subparser for ddb_asr command.
    p_ddb_asr = subparsers.add_parser('ddb_asr', parents=[copts_parser, pandas_parser, slide_parser, plotly_parser],
                                      help=abiview_ddb_asr.__doc__)

    # Subparser for ddb_dipdip command.
    p_ddb_dipdip = subparsers.add_parser('ddb_dipdip', parents=[copts_parser, pandas_parser, slide_parser, plotly_parser],
                                         help=abiview_ddb_dipdip.__doc__)

    # Subparser for ddb_quad command.
    p_ddb_quad = subparsers.add_parser('ddb_quad', parents=[copts_parser, pandas_parser, slide_parser, plotly_parser],
                                       help=abiview_ddb_quad.__doc__)

    # Subparser for ddb_ph_isodistort command.
    p_ddb_isodistort_ph = subparsers.add_parser('ddb_isodistort_ph', parents=[copts_parser],
                                     help=abiview_ddb_isodistort_ph.__doc__)
    p_ddb_isodistort_ph.add_argument("-q", "--qpoint", nargs=3, type=float,
        help="q-point in reduced coordinates e.g. 0.25 0 0. Default: 0, 0, 0", default=[0, 0, 0])

    # Subparser for ddb_ifc command.
    p_ddb_ifc = subparsers.add_parser('ddb_ifc', parents=[copts_parser, pandas_parser, slide_parser],
                                      help=abiview_ddb_ifc.__doc__)

    # Subparser for ddb_qpt command.
    p_ddb_qpt = subparsers.add_parser('ddb_qpt', parents=[copts_parser, pandas_parser, slide_parser],
                                      help=abiview_ddb_qpt.__doc__)

    p_ddb_qpt.add_argument("-q", "--qpoint", nargs=3, type=float,
        help="q-point in reduced coordinates e.g. `-q 0.25 0 0`. Default: 0, 0, 0", default=[0, 0, 0])

    # Subparser for phbands command.
    p_phbands = subparsers.add_parser('phbands', parents=[copts_parser, slide_parser], help=abiview_phbands.__doc__)
    add_args(p_phbands, "xmgrace", "phononweb", "browser", "force")

    # Subparser for lobster command.
    p_lobster = subparsers.add_parser('lobster', parents=[copts_parser, ipy_parser, nb_parser],
        help=abiview_lobster.__doc__)
    p_lobster.add_argument("--prefix", type=str, default="", help="Prefix for lobster output files. Default: ''")

    # Subparser for denpot command.
    #p_denpot = subparsers.add_parser('denpot', parents=[copts_parser], help=abiview_denpot.__doc__)
    #p_denpot.add_argument("-a", "--appname", nargs="?", default=None, const="vesta",
    #        help=("Application name. Default: vesta. "
    #              "Possible options: `%s`, `mayavi`, `vtk`" % ", ".join(Visualizer.all_visunames())))
    #p_denpot.add_argument("--chgcar", default=False, action="store_true", "Convert Density to CHGCAR format.")
    #p_denpot.add_argument("--cube", default=False, action="store_true", "Convert Density/Potential to CUBE format.")

    return parser


@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg:
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse the command line.
    try:
        options = parser.parse_args()
    except Exception:
        show_examples_and_exit(error_code=1)

    if not options.command:
        show_examples_and_exit(error_code=1)

    # loglevel is bound to the string value obtained from the command line argument.
    # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
    import logging
    numeric_level = getattr(logging, options.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % options.loglevel)
    logging.basicConfig(level=numeric_level)

    if getattr(options, "plotly", False): options.expose = True
    if options.verbose > 2: print(options)

    if options.mpl_backend is not None:
        # Set matplotlib backend
        import matplotlib
        matplotlib.use(options.mpl_backend)

    if options.seaborn:
        # Use seaborn settings.
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)

    # Dispatch
    return globals()["abiview_" + options.command](options)


if __name__ == "__main__":
    sys.exit(main())



import os
import wx
import time
import fnmatch

import abipy.gui.awx as awx
import wx.lib.agw.flatnotebook as fnb

from collections import OrderedDict
from monty.dev import get_ncpus
from abipy.gui.events import AbinitEventsFrame, AbinitEventsNotebookFrame
from abipy.gui.timer import MultiTimerFrame, AbinitTimerFrame
from abipy.gui.browser import FileListFrame, DirBrowserFrame, frame_from_filepath, frameclass_from_filepath
from abipy.gui.editor import TextNotebookFrame, SimpleTextViewer
import abipy.flowtk as flowtk


def yaml_manager_dialog(parent):
    """
    Open a dialog that allows the user to select a YAML file with the taskmanager.
    Returns the new manager or None if the user clicked CANCEL or the specifed file is not valid.
    """
    dialog = wx.FileDialog(parent, message="Choose a taskmanager.yml file", defaultDir=os.getcwd(),
                           wildcard="YAML files (*.yml)|*.yml",
                           style=wx.OPEN | wx.MULTIPLE | wx.CHANGE_DIR)

    # Show the dialog and retrieve the user response.
    # If it is the OK response, process the data.
    if dialog.ShowModal() == wx.ID_CANCEL: return None
    filepath = dialog.GetPath()
    dialog.Destroy()

    try:
        return flowtk.TaskManager.from_file(filepath)
    except:
        awx.showErrorMessage(parent)
        return None


class FlowViewerFrame(awx.Frame):
    VERSION = "0.1"

    # Time in second after which we check the status of the tasks.
    REFRESH_INTERVAL = 120

    HELP_MSG = """\
Quick help:

  Task list:

     Left-Click:   Open directory with output files.
     Right-Click:  display popup menu with choices.

Also, these key bindings can be used
(For Mac OSX, replace 'Ctrl' with 'Apple'):

  Ctrl-Q:     quit"""

    def __init__(self, parent, flow, **kwargs):
        """
        Args:
            parent:
                Parent window.
            flow:
                `AbinitFlow` with the list of `Workflow` objects.
        """
        if "title" not in kwargs:
            kwargs["title"] = self.codename

        super(FlowViewerFrame, self).__init__(parent, -1, **kwargs)

        # This combination of options for config seems to work on my Mac.
        self.config = wx.FileConfig(appName=self.codename, localFilename=self.codename + ".ini",
                                    style=wx.CONFIG_USE_LOCAL_FILE)

        # Build menu, toolbar and status bar.
        self.SetMenuBar(self.makeMenu())
        self.makeToolBar()
        self.statusbar = self.CreateStatusBar()
        self.Centre()

        self.flow = flow

        # Disable launch mode if we already executing the flow with the scheduler.
        self.check_launcher_file()

        # Build UI
        self.panel = panel = wx.Panel(self, -1)
        self.main_sizer = main_sizer = wx.BoxSizer(wx.VERTICAL)

        # Here we create a panel and a notebook on the panel
        self.notebook = FlowNotebook(panel, self.flow)
        main_sizer.Add(self.notebook, 1, wx.EXPAND, 5)

        submit_button = wx.Button(panel, -1, label='Submit')
        submit_button.Bind(wx.EVT_BUTTON, self.OnSubmitButton)

        text = wx.StaticText(panel, -1, "Max nlaunch:")
        text.Wrap(-1)
        text.SetToolTipString("Maximum number of tasks that can be submitted. Use -1 for unlimited launches.")
        self.max_nlaunch = wx.SpinCtrl(panel, -1, value=str(get_ncpus()), min=-1)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        hsizer.Add(submit_button, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        hsizer.Add(text, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        hsizer.Add(self.max_nlaunch, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        main_sizer.Add(hsizer, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        panel.SetSizerAndFit(main_sizer)

        # Register this event when the GUI is IDLE
        # Not used anymore since the Yaml parser
        # is very slow when we use the pure python version.
        self.last_refresh = time.time()
        #self.Bind(wx.EVT_IDLE, self.OnIdle)

    @property
    def codename(self):
        """String with the code name."""
        return "Flow Viewer"

    def makeToolBar(self):
        """Create toolbar."""
        self.toolbar = toolbar = self.CreateToolBar()
        toolbar.SetToolBitmapSize(wx.Size(48, 48))

        def bitmap(path):
            return wx.Bitmap(awx.path_img(path))

        self.ID_SHOW_INPUTS = wx.NewId()
        self.ID_SHOW_OUTPUTS = wx.NewId()
        self.ID_SHOW_LOGS = wx.NewId()
        self.ID_SHOW_JOB_SCRIPTS = wx.NewId()
        self.ID_SHOW_JOB_OUTERRS = wx.NewId()
        self.ID_BROWSE = wx.NewId()
        self.ID_SHOW_MAIN_EVENTS = wx.NewId()
        self.ID_SHOW_LOG_EVENTS = wx.NewId()
        self.ID_SHOW_TIMERS = wx.NewId()
        self.ID_CHECK_STATUS = wx.NewId()

        toolbar.AddSimpleTool(self.ID_SHOW_INPUTS, bitmap("in.png"), "Visualize the input file(s) of the workflow.")
        toolbar.AddSimpleTool(self.ID_SHOW_OUTPUTS, bitmap("out.png"), "Visualize the output file(s) of the workflow..")
        toolbar.AddSimpleTool(self.ID_SHOW_LOGS, bitmap("log.png"), "Visualize the log file(s) of the workflow.")
        toolbar.AddSimpleTool(self.ID_SHOW_JOB_SCRIPTS, bitmap("script.png"), "Visualize the script(s).")
        toolbar.AddSimpleTool(self.ID_SHOW_JOB_OUTERRS, bitmap("script.png"), "Visualize the script(s).")
        toolbar.AddSimpleTool(self.ID_BROWSE, bitmap("browse.png"), "Browse all the files of the workflow.")
        toolbar.AddSimpleTool(self.ID_SHOW_MAIN_EVENTS, bitmap("out_evt.png"), "Show the ABINIT events reported in the main output files.")
        toolbar.AddSimpleTool(self.ID_SHOW_LOG_EVENTS, bitmap("log_evt.png"), "Show the ABINIT events reported in the log files.")
        toolbar.AddSimpleTool(self.ID_SHOW_TIMERS, bitmap("timer.png"), "Show the ABINIT timers in the abo files.")

        toolbar.AddSeparator()

        toolbar.AddSimpleTool(self.ID_CHECK_STATUS, bitmap("refresh.png"), "Check the status of the workflow(s).")

        #toolbar.AddSeparator()
        #self.file_selector = FileSelector(toolbar, self)
        #toolbar.AddControl(control=self.file_selector)

        toolbar.Realize()

        # Associate menu/toolbar items with their handlers.
        menu_handlers = [
            (self.ID_SHOW_INPUTS, self.OnShowInputs),
            (self.ID_SHOW_OUTPUTS, self.OnShowOutputs),
            (self.ID_SHOW_LOGS, self.OnShowLogs),
            (self.ID_SHOW_JOB_SCRIPTS, self.OnShowJobScripts),
            (self.ID_SHOW_JOB_OUTERRS, self.OnShowJobOutErrs),
            (self.ID_BROWSE, self.OnBrowse),
            (self.ID_SHOW_MAIN_EVENTS, self.OnShowMainEvents),
            (self.ID_SHOW_LOG_EVENTS, self.OnShowLogEvents),
            (self.ID_SHOW_TIMERS, self.OnShowTimers),
            (self.ID_CHECK_STATUS, self.OnCheckStatusButton),
        ]

        for combo in menu_handlers:
            mid, handler = combo[:2]
            self.Bind(wx.EVT_MENU, handler, id=mid)

    def makeMenu(self):
        """Make the menu bar."""
        menu_bar = wx.MenuBar()

        file_menu = wx.Menu()
        file_menu.Append(wx.ID_OPEN, "&Open", help="Open an existing flow in a new frame")
        file_menu.Append(wx.ID_CLOSE, "&Close", help="Close the file associated to the active tab")
        file_menu.Append(wx.ID_EXIT, "&Quit", help="Exit the application")

        file_history = self.file_history = wx.FileHistory(8)
        file_history.Load(self.config)
        recent = wx.Menu()
        file_history.UseMenu(recent)
        file_history.AddFilesToMenu()
        file_menu.AppendMenu(-1, "&Recent Files", recent)
        self.Bind(wx.EVT_MENU_RANGE, self.OnFileHistory, id=wx.ID_FILE1, id2=wx.ID_FILE9)
        menu_bar.Append(file_menu, "File")

        flow_menu = wx.Menu()

        #self.ID_FLOW_CHANGE_MANAGER = wx.NewId()
        #flow_menu.Append(self.ID_FLOW_CHANGE_MANAGER, "Change TaskManager", help="")

        self.ID_FLOW_OPEN_OUTFILES = wx.NewId()
        flow_menu.Append(self.ID_FLOW_OPEN_OUTFILES, "Open output files", help="")

        self.ID_FLOW_TREE_VIEW = wx.NewId()
        flow_menu.Append(self.ID_FLOW_TREE_VIEW, "Tree view", help="Tree view of the tasks")

        menu_bar.Append(flow_menu, "Flow")

        help_menu = wx.Menu()
        help_menu.Append(wx.ID_ABOUT, "About " + self.codename, help="Info on the application")
        menu_bar.Append(help_menu, "Help")

        self.ID_HELP_QUICKREF = wx.NewId()
        help_menu.Append(self.ID_HELP_QUICKREF, "Quick Reference ", help="Quick reference for " + self.codename)
        help_menu.Append(wx.ID_ABOUT, "About " + self.codename, help="Info on the application")

        # Associate menu/toolbar items with their handlers.
        menu_handlers = [
            (wx.ID_OPEN, self.OnOpen),
            (wx.ID_CLOSE, self.OnClose),
            (wx.ID_EXIT, self.OnExit),
            (wx.ID_ABOUT, self.OnAboutBox),
            #
            #(self.ID_FLOW_CHANGE_MANAGER, self.onChangeManager),
            (self.ID_FLOW_OPEN_OUTFILES, self.onOpenOutFiles),
            (self.ID_FLOW_TREE_VIEW, self.onTaskTreeView),
            #
            (self.ID_HELP_QUICKREF, self.onQuickRef),
        ]

        for combo in menu_handlers:
            mid, handler = combo[:2]
            self.Bind(wx.EVT_MENU, handler, id=mid)

        return menu_bar

    def check_launcher_file(self, with_dialog=True):
        """
        Disable the launch button if we have a sheduler running,
        since we don't want to have processes modifying the flow.
        """
        self.disabled_launcher = False
        pid_files = fnmatch.filter(os.listdir(self.flow.workdir), "*.pid")

        if pid_files:
            pid_file = os.path.join(self.flow.workdir, pid_files[0])
            if not os.path.exists(pid_file): return
            self.disabled_launcher = True

            with open(pid_file, "r") as fh:
                pid = int(fh.readline())

            message = ("Found pid file %s associated to an already running scheduler with pid %d. "
                       "Job submission has been disabled." % (pid_file, pid))

            if with_dialog:
                dialog = wx.MessageDialog(None, message=message, caption='Flow is being executed by a scheduler',
                                          style=wx.OK | wx.ICON_EXCLAMATION)
                dialog.ShowModal()
                dialog.Destroy()

            else:
                self.statusbar.PushStatusText(message)

    def OnSubmitButton(self, event):
        """Submit up to max_nlauch tasks."""
        if self.disabled_launcher: return
        max_nlaunch = int(self.max_nlaunch.GetValue())
        nlaunch = flowtk.PyLauncher(self.flow).rapidfire(max_nlaunch=max_nlaunch)

        self.statusbar.PushStatusText("Submitted %d tasks" % nlaunch)

    def GetSelectedWork(self):
        """
        Return the selected workflow namely that the
        workflow associated to the active tab. None if list is empty.
        """
        return self.notebook.GetSelectedWork()

    def OnIdle(self, event):
        """Function executed when the GUI is idle."""
        now = time.time()
        if (now - self.last_refresh) > self.REFRESH_INTERVAL:
            self.check_launcher_file(with_dialog=False)
            self.CheckStatusAndRedraw()
            self.last_refresh = time.time()

    def OnCheckStatusButton(self, event):
        """Callback triggereed by the checkstatus button."""
        self.CheckStatusAndRedraw()

    def CheckStatusAndRedraw(self):
        """Check the status of all the workflows and redraw the panel."""
        self.statusbar.PushStatusText("Checking status...")

        self.flow.check_status()

        # Count the number of tasks with given status.
        counter = self.flow.status_counter

        # Save the active tab so that we can set it afterwards.
        old_selection = self.notebook.GetSelection()

        # Build new notebook and redraw the panel
        main_sizer = self.main_sizer
        main_sizer.Hide(0)
        main_sizer.Remove(0)
        new_notebook = FlowNotebook(self, self.flow)
        main_sizer.Insert(0, new_notebook, 1, wx.EXPAND, 5)
        self.notebook = new_notebook

        self.panel.Layout()

        # Reinstate the old selection
        self.notebook.SetSelection(old_selection)

        # Write number of jobs with given status.
        message = ", ".join("%s: %s" % (k, v) for (k, v) in counter.items())
        self.statusbar.PushStatusText(message)

    def read_file(self, filepath):
        self.statusbar.PushStatusText("Reading %s" % filepath)

        try:
            flow = flowtk.AbinitFlow.pickle_load(filepath)
            self.AddFileToHistory(filepath)
            return flow
        except:
            awx.showErrorMessage(self)
            return None

    def OnOpen(self, event):
        dialog = wx.FileDialog(
            self, message="Choose a __workflow__.pickle file", defaultDir=os.getcwd(),
            wildcard="pickle files (*.pickle)|*.pickle",
            style=wx.OPEN | wx.MULTIPLE | wx.CHANGE_DIR)

        # Show the dialog and retrieve the user response.
        # If it is the OK response, process the data.
        if dialog.ShowModal() == wx.ID_CANCEL: return

        filepath = dialog.GetPath()
        dialog.Destroy()

        # Add to the history.
        self.file_history.AddFileToHistory(filepath)
        self.file_history.Save(self.config)
        self.config.Flush()
        flow = self.read_file(filepath)

        if flow is not None:
            # Open new frame.
            FlowViewerFrame(self, flow).Show()

    def AddFileToHistory(self, filepath):
        """Add the absolute filepath to the file history."""
        self.file_history.AddFileToHistory(filepath)
        self.file_history.Save(self.config)
        self.config.Flush()

    def OnFileHistory(self, event):
        fileNum = event.GetId() - wx.ID_FILE1
        filepath = self.file_history.GetHistoryFile(fileNum)
        self.file_history.AddFileToHistory(filepath)
        self.read_file(filepath)

    def OnClose(self, event):
        print("onclose")
        #self.flow.pickle_dump()
        #self.Close()

    def OnExit(self, event):
        """Save the status of the flow in the database and exit the GUI."""
        print("onexit")
        #self.flow.pickle_dump()
        self.Close()

    def OnAboutBox(self, event):
        """"Info on the application."""
        awx.makeAboutBox(codename=self.codename, version=self.VERSION,
                         description="", developers="M. Giantomassi")

    def onQuickRef(self, event=None):
        dialog = wx.MessageDialog(self, self.HELP_MSG, self.codename + " Quick Reference",
                               wx.OK | wx.ICON_INFORMATION)
        dialog.ShowModal()
        dialog.Destroy()

    def OnShowInputs(self, event):
        """Show all the input files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        TextNotebookFrame.from_files_and_dir(self, dirpath=work.workdir, walk=True, wildcard="*.abi").Show()

    def OnShowOutputs(self, event):
        """Show all the output files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        TextNotebookFrame.from_files_and_dir(self, dirpath=work.workdir, walk=True, wildcard="*.abo").Show()

    def OnShowJobScripts(self, event):
        """Show all the job script files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        TextNotebookFrame.from_files_and_dir(self, dirpath=work.workdir, walk=True, wildcard="*.sh").Show()

    def OnShowJobOutErrs(self, event):
        """Show all the queue output/error files files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        TextNotebookFrame.from_files_and_dir(self, dirpath=work.workdir, walk=True, wildcard="*.qout|*.qerr").Show()

    def OnShowLogs(self, event):
        """Show all the log files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        TextNotebookFrame.from_files_and_dir(self, dirpath=work.workdir, walk=True, wildcard="*.log").Show()

    def OnBrowse(self, event):
        """Browse all the output files produced by the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        FileListFrame(self, dirpaths=work.workdir).Show()
        #DirBrowserFrame(self, dir=work.workdir).Show()

    def OnShowMainEvents(self, event):
        """Browse all the main events of the tasks in the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        AbinitEventsNotebookFrame(self, filenames=[task.output_file.path for task in work]).Show()

    def OnShowLogEvents(self, event):
        """Browse all the log events of the tasks in the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        AbinitEventsNotebookFrame(self, [task.log_file.path for task in work]).Show()

    def OnShowTimers(self, event):
        """Analyze the timing data of all the output files of the selected `Workflow`."""
        work = self.GetSelectedWork()
        if work is None: return
        timers = work.parse_timers()
        # Build the frame for analyzing multiple timers.
        MultiTimerFrame(self, timers).Show()

    def onTaskTreeView(self, event):
        TaskTreeView(self, self.flow).Show()

    #def onChangeManager(self, event):
    #    #ChangeTaskManager(self, self.flow).Show()
    #    new_manager = yaml_manager_dialog(self)
    #    if new_manager is None: return
    #    print(new_manager)
    #    #status_selected =  upper()
    #    #status = None if status_selected == "ALL" else status_selected
    #    # Change the manager of the errored tasks.
    #    #for task in flow.iflat_tasks(status="S_ERROR"):
    #    #    task.reset()
    #    #    task.set_manager(new_manager)

    def onOpenOutFiles(self, event):
        FileSelectorFrame(parent=self, viewer=self).Show()


class FlowNotebook(fnb.FlatNotebook):
    """
    Notebook class
    """
    def __init__(self, parent, flow):
        try:
            style = fnb.FNB_NO_X_BUTTON | fnb.FNB_NAV_BUTTONS_WHEN_NEEDED
        except AttributeError:
            style = fnb.FNB_NO_X_BUTTON

        super(FlowNotebook, self).__init__(parent, id=-1, style=style)

        self.flow = flow
        for work in flow:
            tab = TabPanel(self, work)
            self.AddPage(tab, text=os.path.basename(work.workdir))

    def GetSelectedWork(self):
        """
        Return the selected workflow namely that the workflow associated to the
        active tab. None is list is empy.
        """
        # FIXME: If we want to allow the user to remove some tab we have
        # to remove the corresponding work from workflows.
        # Easy if workflow_viewer can only read data but it might be source
        # of bugs if we want to modify the object e.g. by submitting the calculation.
        # For the time-being, use this assertion to prevent users from removing pages.
        if self.GetPageCount() != len(self.flow):
            return awx.showErrorMessage(self, message="Bad user has removed pages from the notebook!")

        idx = self.GetSelection()
        if idx == -1: return None
        try:
            return self.flow[idx]
        except IndexError:
            return None


class TabPanel(wx.Panel):
    """
    Notebook tab.
    """
    def __init__(self, parent, work, **kwargs):
        wx.Panel.__init__(self, parent=parent, id=-1, **kwargs)

        # List Control with the individual tasks of the workflow.
        task_listctrl = TaskListCtrl(self, work)

        label = wx.StaticText(self, -1, "Workflow class %s, status: %s, finalized: %s" % (
            work.__class__.__name__, work.status, work.finalized))
        label.Wrap(-1)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(label, 0, wx.ALIGN_LEFT, 5)
        main_sizer.Add(task_listctrl, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(main_sizer)


class TaskListCtrl(wx.ListCtrl):
    """
    ListCtrl with the list of task in the `Workflow`.
    """
    def __init__(self, parent, work, **kwargs):
        """
        Args:
            parent:
                Parent window.
            work:
                `Workflow` containig the List of `Task` instances.
        """
        super(TaskListCtrl, self).__init__(parent, id=-1, style=wx.LC_REPORT | wx.BORDER_SUNKEN, **kwargs)

        self.work = work

        columns = ["Task", "Status", "Queue_id",
                   "Errors", "Warnings", "Comments",
                   "MPI", "OMP",
                   "num_restarts", "Task Class",
                   ]

        for index, col in enumerate(columns):
            self.InsertColumn(index, col)

        # Used to store the Max width in pixels for the data in the column.
        column_widths = [awx.get_width_height(self, s)[0] for s in columns]

        for task in work:
            events = map(str, 3*["N/A"])

            try:
                report = task.get_event_report()
                if report is not None:
                    events = map(str, [report.num_errors, report.num_warnings, report.num_comments])
            except:
                pass

            cpu_info = [task.mpi_procs, task.omp_threads]
            entry = map(str, [task.name, str(task.status), task.queue_id] +
                              events +
                              cpu_info +
                              [task.num_restarts, task.__class__.__name__]
                        )
            w = [awx.get_width_height(self, s)[0] for s in entry]
            column_widths = map(max, zip(w, column_widths))

            self.Append(entry)

        # Set the width in pixel for each column.
        for index, col in enumerate(columns):
            self.SetColumnWidth(index, column_widths[index])

        self.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self.OnRightClick)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated)

    def OnRightClick(self, event):
        currentItem = event.m_itemIndex
        if currentItem == -1: return

        # Open the popup menu then destroy it to avoid mem leak.
        task = self.work[currentItem]
        menu = TaskPopupMenu(parent=self, task=task)
        self.PopupMenu(menu, event.GetPoint())
        menu.Destroy()

    def OnItemActivated(self, event):
        """
        Browse the outdir of the task
        """
        currentItem = event.m_itemIndex
        task = self.work[currentItem]
        FileListFrame(self, dirpaths=task.outdir.path, title=task.outdir.relpath).Show()


# Callbacks

def show_task_main_output(parent, task):
    """Show the main output file of the task."""
    file = task.output_file

    if file.exists:
        frame_from_filepath(parent, file.path).Show()
    else:
        awx.showErrorMessage(parent=parent, message="Output file %s does not exist" % file.path)


def show_task_log(parent, task):
    """Show the log file of the task."""
    file = task.log_file

    if file.exists:
        frame_from_filepath(parent, file.path).Show()
    else:
        awx.showErrorMessage(parent=parent, message="Output file %s does not exist" % file.path)


def show_task_main_events(parent, task):
    """Show the events reported in the main output file."""
    file = task.output_file

    if file.exists:
        AbinitEventsFrame(parent, file.path).Show()
    else:
        awx.showErrorMessage(parent=parent, message="Output file %s does not exist" % file.path)


def show_task_log_events(parent, task):
    """Show the events reported in the log file."""
    file = task.log_file

    if file.exists:
        AbinitEventsFrame(parent, file.path).Show()
    else:
        awx.showErrorMessage(parent=parent, message="Log file %s does not exist" % file.path)


def browse_indir(parent, task):
    """Open a window that allows the user to browse the input files in indir."""
    FileListFrame(parent, dirpaths=task.indir.path).Show()


def browse_outdir(parent, task):
    """Open a window that allows the user to browse the output files in outdir."""
    FileListFrame(parent, dirpaths=task.outdir.path).Show()


def browse_tmpdir(parent, task):
    """Open a window that allows the user to browse the output files in outdir."""
    FileListFrame(parent, dirpaths=task.tmpdir.path).Show()


def show_history(parent, task):
    """Show the history of the task."""
    text = "\n".join(task.history)
    SimpleTextViewer(parent, text).Show()


def show_timer(parent, task):
    """Show timing data of the k."""
    try:
        frame = AbinitTimerFrame(parent, task.output_file.path)
        frame.Show()
    except awx.Error as exc:
        awx.showErrorMessage(parent, str(exc))


def check_status_and_pickle(task):
    """Check the status of the task and update the pickle database."""
    task.flow.check_status()
    task.flow.pickle_dump()


def task_restart(parent, task):
    """Restart the task."""
    task.restart()
    check_status_and_pickle(task)


def task_reset(parent, task):
    """Reset the status of the task."""
    task.reset()
    check_status_and_pickle(task)


def task_set_status(parent, task):
    """Reset the status of the task."""
    choices = [str(s) for s in flowtk.Node.ALL_STATUS]
    dialog = wx.SingleChoiceDialog(parent, message="Select new status", caption="", choices=choices)
    if dialog.ShowModal() == wx.ID_CANCEL: return None
    status = choices[dialog.GetSelection()]
    dialog.Destroy()

    task.set_status(status, info_msg="Status changed by user on %s" % time.asctime())
    #task.reset()
    check_status_and_pickle(task)


def task_show_deps(parent, task):
    """Show the dependencies of the task."""
    text = task.str_deps()
    SimpleTextViewer(parent, text).Show()


def task_inspect(parent, task):
    """Inspect the results at runtime."""
    if hasattr(task, "inspect"):
        task.inspect()


class TaskPopupMenu(wx.Menu):
    """
    A `TaskPopupMenu` has a list of callback functions indexed by the menu title.
    The signature of the callback function is func(parent, task) where parent is
    the wx Window that will become the parent of the new frame created by the callback.
    and task is a `Task` instance.
    """
    MENU_TITLES = OrderedDict([
        ("output", show_task_main_output),
        ("log", show_task_log),
        ("main events", show_task_main_events),
        ("log events",  show_task_log_events),
        ("browse indir", browse_indir),
        ("browse outdir", browse_outdir),
        ("browse tmpdir", browse_tmpdir),
        ("history", show_history),
        ("restart", task_restart),
        ("reset", task_reset),
        ("dependencies", task_show_deps),
        ("inspect", task_inspect),
        ("set status", task_set_status),
        ("timer", show_timer),
    ])

    def __init__(self, parent, task):
        super(TaskPopupMenu, self).__init__()
        self.parent, self.task = parent, task

        self._make_menu()

    def _make_menu(self):
        """Build the menu"""
        self.menu_title_by_id = OrderedDict()

        for title in self.MENU_TITLES:
            self.menu_title_by_id[wx.NewId()] = title

        for (id, title) in self.menu_title_by_id.items():
            # Register menu handlers with EVT_MENU, on the menu.
            self.Append(id, title)
            wx.EVT_MENU(self, id, self.OnMenuSelection)

    def _get_callback(self, title):
        return self.MENU_TITLES[title]

    def OnMenuSelection(self, event):
        title = self.menu_title_by_id[event.GetId()]
        callback = self._get_callback(title)
        #print("Calling callback %s on task %s" % (callback, self.task))
        try:
            callback(self.parent, self.task)
        except:
            awx.showErrorMessage(parent=self.parent)

#class ChangeTaskManager(awx.Frame):
#    def __init__(self, parent, flow, **kwargs)
#        super(ChangeTaskManager, self).__init__(self, parent, **kwargs)
#
#
#    def onOkButton(self, event):
#        new_manager = yaml_manager_dialog(self)
#        if new_manager is None: return
#        print(new_manager)
#        #status_selected =  upper()
#        #status = None if status_selected == "ALL" else status_selected
#        # Change the manager of the errored tasks.
#        #for task in flow.iflat_tasks(status="S_ERROR"):
#        #    task.reset()
#        #    task.set_manager(new_manager)


class TaskStatusTreePanel(awx.Panel):
    """
    Panel with a TreeCtrl that allows the user to navigate the tasks grouped by status.
    """
    def __init__(self, parent, flow, **kwargs):
        super(TaskStatusTreePanel, self).__init__(parent, -1, **kwargs)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        vbox = wx.BoxSizer(wx.VERTICAL)
        panel1 = awx.Panel(self, -1)
        panel2 = awx.Panel(self, -1)

        self.tree = tree = wx.TreeCtrl(panel1, 1, wx.DefaultPosition, (-1, -1), wx.TR_HIDE_ROOT | wx.TR_HAS_BUTTONS)

        root = self.tree.AddRoot('Task Status')

        # entry = collections.namedtuple("Entry", "task wi ti")
        #print(status2entries)
        self.status2entries = flow.groupby_status()

        self.status_nodes = []
        for status, entries in self.status2entries.items():
            node = tree.AppendItem(root, "%d %s tasks" % (len(entries), str(status)), data=wx.TreeItemData(status))
            self.status_nodes.append(node)
            for entry in entries:
                tree.AppendItem(node, "Task: " + str(entry.task), data=wx.TreeItemData(entry))

        tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.onSelChanged)
        tree.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.onItemRightClick)

        self.display = wx.StaticText(panel2, -1, '', (10, 10), style=wx.ALIGN_LEFT)

        vbox.Add(self.tree, 1, wx.EXPAND)
        main_sizer.Add(panel1, 1, wx.EXPAND)
        main_sizer.Add(panel2, 1, wx.EXPAND)
        panel1.SetSizerAndFit(vbox)

        self.SetSizerAndFit(main_sizer)
        self.Centre()

    def onSelChanged(self, event):
        node = event.GetItem()
        if node in self.status_nodes: return

        proxy = self.tree.GetItemData(node)
        if proxy is None: return
        entry = proxy.GetData()

        task = entry.task
        s = str(task)
        s += "\nHistory:\n" + task.str_history

        self.display.SetLabel(s)

    def onItemRightClick(self, event):
        node = event.GetItem()
        proxy = self.tree.GetItemData(node)
        if proxy is None: return

        if node in self.status_nodes:
            status = proxy.GetData()
            print("received set of tasks with status %s" % status)
            popup_menu = self.makePopupMenuStatus()
            self.PopupMenu(popup_menu, event.GetPoint())
            popup_menu.Destroy()

        #print("click")
        #print("event", dir(event))

    def makePopupMenuStatus(self):
        self.ID_POPUP_CHANGE_MANAGER = wx.NewId()
        menu = wx.Menu()
        menu.Append(self.ID_POPUP_CHANGE_MANAGER, "Change manager")

        # Associate menu/toolbar items with their handlers.
        menu_handlers = [
            (self.ID_POPUP_CHANGE_MANAGER, self.onChangeManager),
        ]

        for combo in menu_handlers:
            mid, handler = combo[:2]
            self.Bind(wx.EVT_MENU, handler, id=mid)

        return menu

    def onChangeManager(self, event):
        print("changer manager")
        node = self.tree.GetSelection()
        status = self.tree.GetItemData(node).GetData()
        print("status", status)

        entries = self.status2entries[status]

        new_manager = yaml_manager_dialog(self)
        for e in entries:
            e.task.reset()
            e.task.set_manager(new_manager)

        #self.flow.build_and_pickle_dump()


class TaskTreeView(awx.Frame):
    """
    A frame with a TaskStatusTreePanel.
    """
    def __init__(self, parent, flow, **kwargs):
        if "title" not in kwargs:
            kwargs["title"] = "Task tree view: %s" % flow.workdir

        super(TaskTreeView, self).__init__(parent, **kwargs)

        panel = TaskStatusTreePanel(self, flow)


class FileSelector(awx.Panel):
#class FileSelector(wx.Control):
    """
    Control that allows the user to select multiple output files of the same type (either inside
    a `Workflow` on in the entire `Flow`. The open button will open a viewer to analyze
    the multiple files selected.
    """
    def __init__(self, parent, viewer, **kwargs):
        super(FileSelector, self).__init__(parent, -1, **kwargs)
        self.viewer = viewer
        panel = self #wx.Panel(self, -1)

        wcards = ["*GSR.nc", "*WFK-etsf.nc", "*SIGRES.nc", "*MDF.nc"]

        self.wcards_cbox = wx.ComboBox(panel, id=-1, name='File type', choices=wcards, value=wcards[0], style=wx.CB_READONLY)

        smodes = ["Selected Workflow", "Entire Flow"]
        self.select_rbox = wx.RadioBox(panel, id=1, name="Selection Mode", choices=smodes, style=wx.RA_SPECIFY_ROWS)

        open_button = wx.Button(panel, -1, label='Open files')
        open_button.Bind(wx.EVT_BUTTON, self.onOpenButton)

        #main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        main_sizer = wx.GridBagSizer(10, 10)

        #vsizer = wx.BoxSizer(wx.VERTICAL)
        #vsizer.Add(self.wcards_cbox, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        #vsizer.Add(open_button, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        #main_sizer.Add(vsizer)
        #main_sizer.Add(self.select_rbox, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        main_sizer.Add(self.wcards_cbox, (0, 0), (1,1), wx.ALIGN_CENTER)
        main_sizer.Add(open_button, (1, 0), (1,1), wx.ALIGN_CENTER)
        main_sizer.Add(self.select_rbox, (0, 1), (3,3), wx.EXPAND)

        panel.SetSizerAndFit(main_sizer)

    def getWildCard(self):
        """Returns a string with the Abinit extension selected by the user."""
        return self.wcards_cbox.GetValue()

    def getSelectionMode(self):
        """Returns a string with the selection mode selected by the user."""
        index = self.select_rbox.GetSelection()
        return self.select_rbox.GetString(index)

    def onOpenButton(self, event):
        wcard = self.getWildCard()
        smode = self.getSelectionMode()
        print("wcard", wcard, "smode", smode)

        # Find the filepaths according to smode.
        filepaths = []

        if smode == "Selected Workflow":
            work = self.viewer.GetSelectedWork()
            for task in work:
                filepaths.extend(task.outdir.list_filepaths(wildcard=wcard))

        elif smode == "Entire Flow":
            for work in self.viewer.flow:
                for task in work:
                    filepaths.extend(task.outdir.list_filepaths(wildcard=wcard))

        else:
            return awx.showErrorMessage(self, "Wrong value of smode: %s" % smode)

        if not filepaths:
            return awx.showErrorMessage(self, "Cannot find any file matching the specified shell pattern")

        print("filepaths", filepaths)

        # Get the viewer class associated to these files, build the frame and show it.
        frame_class = frameclass_from_filepath(filepaths[0])
        if frame_class is None: return
        frame_class(self, filepaths).Show()


class FileSelectorFrame(wx.Frame):
    def __init__(self, parent, viewer, **kwargs):
        super(FileSelectorFrame, self).__init__(parent, -1, **kwargs)

        panel = wx.Panel(self, -1)
        file_selector = FileSelector(panel, viewer)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(file_selector, 1, wx.EXPAND , 5)
        panel.SetSizerAndFit(main_sizer)


#class StatusSelectorFrame(wx.Frame):
#    def __init__(self, parent, viewer, **kwargs):
#        super(StatusSelectorFrame, self).__init__(parent, -1, **kwargs)
#
#        panel = wx.Panel(self, -1)
#
#        #choices = ["S_OK", "*WFK-etsf.nc", "*SIGRES.nc", "*MDF.nc"]
#
#        #self.status_cbox = wx.ComboBox(panel, id=-1, name='File type', choices=choices, value=choices[0], style=wx.CB_READONLY)
#
#        #smodes = ["Selected Workflow", "Entire Flow"]
#        #self.select_rbox = wx.RadioBox(panel, id=1, name="Selection Mode", choices=smodes, style=wx.RA_SPECIFY_ROWS)
#
#        #open_button = wx.Button(panel, -1, label='Open files')
#        #open_button.Bind(wx.EVT_BUTTON, self.onOpenButton)
#
#        main_sizer = wx.BoxSizer(wx.VERTICAL)
#        main_sizer.Add(file_selector, 1, wx.EXPAND , 5)
#        panel.SetSizerAndFit(main_sizer)
#
#    def getSelectedStatus(self):
#        return self.choices_cbox.GetValue()


def wxapp_flow_viewer(works):
    """Standalone application for `FlowViewerFrame`."""
    app = awx.App()
    FlowViewerFrame(None, works).Show()
    return app



# coding: utf-8
"""IO related utilities."""
from __future__ import annotations

import os
import ruamel.yaml as yaml

from contextlib import ExitStack
from subprocess import call
from typing import Any
from monty.termcolor import cprint
from monty.string import list_strings


def yaml_safe_load(string: str) -> Any:
    """Load Yaml string"""
    return yaml.YAML(typ='safe', pure=True).load(string)


def yaml_safe_load_path(filepath: str) -> Any:
    """Load Yaml document from filepath"""
    with open(filepath, "rt") as fh:
        return yaml.YAML(typ='safe', pure=True).load(fh.read())


class ExitStackWithFiles(ExitStack):
    """
    Context manager for dynamic management of a stack of file-like objects.
    Mainly used in a callee that needs to return files to the caller

    Usage example:

    .. code-block:: python

        exit_stack = ExitStackWithFiles()
        exit_stack.enter_context(phbst_file)
        return exit_stack
    """
    def __init__(self):
        self.files = []
        super().__init__()

    def enter_context(self, myfile):
        # If my file is None, we add it to files but without registering the callback.
        self.files.append(myfile)
        if myfile is not None:
            return super().enter_context(myfile)

    def __iter__(self):
        return self.files.__iter__()

    def __next__(self):
        return self.files.__next__()

    def __getitem__(self, slice):
        return self.files.__getitem__(slice)


def get_input(prompt):
    """
    Wraps python builtin input so that we can easily mock it in unit tests using:

        from unittest.mock import patch
        with patch('abipy.tools.iotools.get_input', return_value='no'):
            do_something_that_uses_get_input
    """
    return input(prompt)


def ask_yes_no(prompt: str, default=None):  # pragma: no cover
    """
    Ask a question and return a boolean (y/n) answer.

    If default is given (one of 'y','n'), it is used if the user input is
    empty. Otherwise the question is repeated until an answer is given.

    An EOF is treated as the default answer.  If there is no default, an
    exception is raised to prevent infinite loops.

    Valid answers are: y/yes/n/no (match is not case sensitive).
    """
    # Fixes py2.x
    answers = {'y': True, 'n': False, 'yes': True, 'no': False}
    ans = None
    while ans not in answers.keys():
        try:
            ans = get_input(prompt + ' ').lower()
            if not ans:
                # response was an empty string
                ans = default
        except KeyboardInterrupt:
            pass
        except EOFError:
            if default in answers.keys():
                ans = default
                print("")
            else:
                raise

    return answers[ans]


def _user_wants_to_exit(): # pragma: no cover
    try:
        answer = get_input("Do you want to continue [Y/n]")
    except EOFError:
        return True

    if answer.lower().strip() in ["n", "no"]: return True
    return False


class EditorError(Exception):
    """Base class for exceptions raised by `Editor`"""


class Editor(object):  # pragma: no cover
    DEFAULT_EDITOR = "vi"

    Error = EditorError

    def __init__(self, editor=None):
        if editor is None:
            self.editor = os.getenv("EDITOR", self.DEFAULT_EDITOR)
        else:
            self.editor = str(editor)

    def edit_file(self, fname):
        retcode = call([self.editor, fname])
        if retcode != 0:
            cprint("Retcode %s while editing file: %s" % (retcode, fname), "red")
        return retcode

    def edit_files(self, fnames, ask_for_exit=True):
        for idx, fname in enumerate(list_strings(fnames)):
            exit_status = self.edit_file(fname)

            if exit_status != 0:
                return exit_status

            if ask_for_exit and idx != len(fnames) - 1 and _user_wants_to_exit():
                break

        return 0


def input_from_editor(message=None):  # pragma: no cover
    if message is not None:
        print(message, end="")

    from tempfile import mkstemp
    fd, fname = mkstemp(text=True)

    Editor().edit_file(fname)
    with open(fname, "rt") as fileobj:
        return fileobj.read()
